/*
 * recompose - combine component parts into a firmware file
 *
 * Copyright (C) 2004 Robert Bowler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* I have not given a seconds thought to how this will work on big endian machines */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"

int reverse_endian = 1;

void usage(void)
{
  fprintf(stderr, "usage: recompose [-v version] [-m model]");
  exit(1);
}

void write_component(char *buf,
		     int len,
		     const char *filename)
{
  
 if (len > 0)
   {
     FILE *f = fopen(filename, "w");
     fwrite(buf, len, 1, f);
     fclose(f);
   }
}

void
read_component(char *buf,
	       int length,
	       const char *filename)
{
  if (length > 0)
    {
      FILE *component = fopen(filename, "r");
      if (component == NULL)
	{
	  fprintf(stderr, "failed to open %s\n", filename);
	  exit(1);
	}
      fread(buf, length, 1, component);
      fclose(component);
      printf("read in %s\n", filename);
    }
}



int
main(int argc,
	char *argv[])
{
  FILE *bin_file;
  struct stat buf;
  unsigned char *file_buffer = NULL;
  struct header_type *header;
  unsigned image_size, loader_size, root_fs_size, kernel_size;
  unsigned loader_addr, root_fs_addr, kernel_addr;
  char vendor[21] = "Broadcom Corporation";
  char version[21] = "Firmware version 1.0";
  char model[17] = "96345R";  
  char header_version = 0;
  int c;

  if (0 == stat("header", &buf)
      && buf.st_size == 256)
    {
      struct header_type tmp_header;
      FILE *header = fopen("header", "r");
      if (NULL != header)
	{
	  printf("Defaulting header strings from existing header\n");
	  fread(&tmp_header, 256, 1, header);

	  memcpy(vendor, tmp_header.vendor, 20);
	  vendor[20] = 0;
	  memcpy(version, tmp_header.version, 20);
	  version[20] = 0;
	  memcpy(model, tmp_header.model, 16);
	  model[16] = 0;
	  header_version = tmp_header.header_version;
	  fclose (header);
	}
    }

  opterr = 0;
  while ((c = getopt (argc, argv, "m:")) != -1)
    {
      switch (c)
	{
	case 'z':
	  strncpy(vendor, optarg, 20);
	  vendor[20] = 0;
	  break;
	case 'v':
	  strncpy(version, optarg, 20);
	  version[20] = 0;
	  break;
	case 'm':
	  strncpy(model, optarg, 16);
	  model[16] = 0;
	  break;
	default:
	  usage();
	}     
    }

  if (0 == stat("loader", &buf))
    {
      loader_size = buf.st_size;
      loader_addr = 0xBFC00000;
    }
  else
    {
      loader_size = 0;
      loader_addr = 0;
    }
  
  if (0 != stat("root_fs", &buf))
    {
      fprintf(stderr,"Failed to open %s\n", "root_fs");
      exit(1);
    }
  root_fs_size = buf.st_size;
  root_fs_addr = 0xBFC10100;
  
  if (0 != stat("kernel", &buf))
    {
      fprintf(stderr,"Failed to open %s\n", "kernel");
      exit(1);
    }
  kernel_size = buf.st_size;
  kernel_addr = root_fs_addr + root_fs_size;
  
  image_size = loader_size + root_fs_size + kernel_size;
  file_buffer = calloc(1, image_size + 256);
  if (NULL == file_buffer)
    {
      fprintf(stderr, "out of memory\n");
      exit(1);
    }
  read_component(file_buffer + 256, loader_size, "loader");
  read_component(file_buffer + 256 + loader_size, root_fs_size, "root_fs");
  read_component(file_buffer + 256 + loader_size + root_fs_size,
		 kernel_size, "kernel");

  header = file_buffer;
  header->magic = 0x00000036;
  sprintf(&header->vendor, vendor);
  sprintf(&header->version, version);
  sprintf(&header->model, model);  
  header->unknown[0] = 0x31;
  header->unknown[1] = 0x00;
  sprintf(&header->image_size,"%u", image_size);
  sprintf(&header->loader_addr, "%u", loader_addr);
  sprintf(&header->loader_size, "%u", loader_size);
  sprintf(&header->root_fs_addr, "%u", root_fs_addr);
  sprintf(&header->root_fs_size, "%u", root_fs_size);
  sprintf(&header->kernel_addr, "%u", kernel_addr);
  sprintf(&header->kernel_size, "%u", kernel_size);

  if (header_version == '2')
    {
      printf("writing version 2 firmware\n");
      header->header_version = header_version;
      header->payload_checksum = htocl(~crc32(file_buffer + 256 + loader_size
				       , root_fs_size + kernel_size));
    }
  header->data_checksum = htocl(~crc32(file_buffer + 256, image_size));
  header->header_checksum = htocl(~crc32(file_buffer, 236));  

  if (*(unsigned *)(file_buffer + 256 + loader_size) != htocl(0x28cd3d45))
    {
      fprintf(stderr, "bad magic on root_fs - not writing firmware\n");
      exit(1);
    }

  bin_file = fopen("firmware.bin", "w");
  if (bin_file == NULL)
    {
      fprintf(stderr, "failed to open firmware file\n");
      exit(1);
    }      
  fwrite(file_buffer, 1, image_size + 256, bin_file);
  fclose(bin_file);
  return 0;
}
